Directives Glossary
===================

Here we list the common playbook objects and their directives.
Note that not all directives affect the object itself and might just be there to be inherited by other contained objects.
Aliases for the directives are not reflected here, nor are mutable ones, for example `action` in task can be substituted by the name of any module plugin.

.. contents::
   :local:
   :depth: 1

{% for name in clist %}

{{ name }}
{{ '-' * name|length }}
{% for attribute in oblist[name]|sort %}
  * {{ attribute }}
{% endfor %}

{% endfor %}
